import inspect
import logging
import pathlib
from collections import Callable

from lightning_fast.tools.logging_tools import LoggerUtil
from lightning_fast.tools.meta_tools import SingletonMeta


class LogInjection(metaclass=SingletonMeta):
    """
    实例化此类然后调用generate_log_injection会生成一个类装饰器,
    使用此装饰器之后会将被装饰的类的__info__方法打印并保存日志,
    被装饰类的__info__方法返回字符串，之后类中所有调用__info__方法的记录都会使用logging打印。
    """

    def __init__(
        self,
        logger_name: str = None,
        logger_level: int = logging.INFO,
        log_dir: pathlib.Path = None,
    ):
        if not logger_name or not log_dir:
            self.info_logger = LoggerUtil.get_simple_logger()
        else:
            self.info_logger = LoggerUtil.get_simple_logger(
                log_dir, f"{logger_name}_info", logger_level
            )
        # self.info_logger.setLevel(logging.INFO)

    def generate_log_injection(self) -> Callable:
        def log_injection(other_cls):
            if hasattr(other_cls, "__info__"):
                origin_info_func = other_cls.__info__

                def __inject_info(*args, **kwargs):

                    if inspect.ismethod(origin_info_func):
                        # 如果是绑定方法即类方法或者普通方法
                        self.info_logger.info(origin_info_func(*args[1:], **kwargs))
                    else:
                        self.info_logger.info(origin_info_func(*args, **kwargs))

                setattr(other_cls, "__info__", __inject_info)
            return other_cls

        return log_injection



from ctypes import (
    POINTER,
    c_void_p,
    c_uint,
    Structure,
    c_uint32,
    sizeof,
    c_long,
    c_longlong,
    c_uint8,
)

if sizeof(c_void_p) == 4:
    # ptrdiff_t = c_int32
    ptrdiff_t = c_long
elif sizeof(c_void_p) == 8:
    # ptrdiff_t = c_int64
    ptrdiff_t = c_longlong
else:
    raise ValueError(f"什么乱遭的系统，指针大小为{sizeof(c_void_p)}")


class UT_hash_table(Structure):
    pass


class UT_hash_handle(Structure):
    pass


UT_hash_handle._fields_ = [
    ("tbl", POINTER(UT_hash_table)),
    ("prev", c_void_p),
    ("next", c_void_p),
    ("hh_prev", POINTER(UT_hash_handle)),
    ("hh_next", POINTER(UT_hash_handle)),
    ("key", c_void_p),
    ("keylen", c_uint),
    ("hashv", c_uint),
]


class UT_hash_bucket(Structure):
    _fields_ = [
        ("hh_head", POINTER(UT_hash_handle)),
        ("count", c_uint),
        ("expand_mult", c_uint),
    ]


UT_hash_table._fields_ = [
    ("buckets", POINTER(UT_hash_bucket)),
    ("num_buckets", c_uint),
    ("log2_num_buckets", c_uint),
    ("num_items", c_uint),
    ("tail", POINTER(UT_hash_handle)),
    ("hho", ptrdiff_t),
    ("ideal_chain_maxlen", c_uint),
    ("nonideal_items", c_uint),
    ("ineff_expands", c_uint),
    ("noexpand", c_uint),
    ("signature", c_uint32),
    ("bloom_sig", c_uint32),
    ("bloom_bv", POINTER(c_uint8)),
    ("bloom_nbits", c_uint8),
]



import pathlib
from typing import Optional

from lightning_fast.tools.path_tools import DirectoryDescriptor


class SimpleDirectoryCollection(dict):
    """
    将config中的pathlib.Path(__file__).absolute传进来，总之这里传一个项目根目录的绝对路径进来
    """

    base_dir = DirectoryDescriptor(None)
    data_dir = DirectoryDescriptor(None)
    log_dir = DirectoryDescriptor(None)
    tmp_dir = DirectoryDescriptor(None)

    def __init__(self, base_directory: Optional[pathlib.Path]):
        if base_directory is not None:
            self.base_dir = base_directory
            self.data_dir = base_directory / "data"
            self.log_dir = base_directory / "log"
            self.tmp_dir = base_directory / "tmp"
        super().__init__()



import string


class FormatTools:
    @classmethod
    def clear_punctuation(cls, sentence):
        """
        此方法用于去除所有标点符号
        :param sentence: 欲处理字符串
        :return: 处理后字符串
        """
        return sentence.translate(str.maketrans("", "", string.punctuation))



import socket


class IpTools:
    @classmethod
    def get_ip(cls):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable, A type private address.
            s.connect(("10.255.255.255", 1))
            ip = s.getsockname()[0]
        except OSError:
            ip = "127.0.0.1"
        finally:
            s.close()
        return ip



from lightning_fast.mongodb.clients.client_descriptor.descriptors.sync_client_descriptor import (
    SyncMongodbClient,
)
from lightning_fast.mongodb.clients.client_descriptor.descriptors.async_client_descriptor import (
    AsyncMongodbClient,
)

__all__ = ["AsyncMongodbClient", "AsyncMongodbClient"]



import multiprocessing
import pathlib
import re
import socket

import dask
import psutil

from dask.distributed import LocalCluster, Client


class DaskConfigController:
    @classmethod
    def set_local_one_thread_config(cls):
        dask.config.set({"scheduler": "single-threaded"})

    @classmethod
    def set_local_cluster_config(
        cls, tmp_directory: pathlib.Path, worker_count=None, dashboard_port=5000
    ):
        if worker_count is None:
            worker_count = (
                16 if multiprocessing.cpu_count() > 16 else multiprocessing.cpu_count()
            )
        else:
            worker_count = worker_count
        valid_memory_mb = psutil.virtual_memory().available >> 20
        memory_limit = int(valid_memory_mb / worker_count)
        dask.config.set(
            {
                "temporary-directory": str(tmp_directory / "dask-worker-space"),
                "memory-limit": f"{memory_limit} MB",
            }
        )
        cluster = LocalCluster(
            n_workers=worker_count, dashboard_address=f":{dashboard_port}"
        )
        Client(cluster)
        try:
            origin_ip = socket.gethostbyname(socket.gethostname())
            # noinspection HttpUrlsUsage
            origin_hostname = re.sub(
                r"^.*:", "http://" + origin_ip + ":", cluster.dashboard_link
            )
            print(
                f"\033[1;32;43m dashboard_address(origin): {origin_hostname} \033[0m!"
            )
            print(
                f"\033[1;32;43m dashboard_address(local): {cluster.dashboard_link} \033[0m!"
            )
        except socket.gaierror:
            print(
                f"\033[1;32;43m dashboard_address(local): {socket.gethostname()} \033[0m!"
            )



import dask.dataframe as dd
from dask.diagnostics import ProgressBar
from dask.distributed import progress


class DaskComputer:
    @classmethod
    def compute_local_one_thread(cls, *args):
        with ProgressBar():
            return dd.compute(*args)

    @classmethod
    def compute_local_cluster(cls, *args):
        targets = [arg.persist() for arg in args]
        progress(*targets)
        return dd.compute(*targets)



import pandas as pd
from dateutil.parser import parse


class SpecialDatePoints(object):
    @classmethod
    def get_month_list(cls, start_year_month: str, end_year_month: str):
        """
        返回左闭右开的所有月份，格式为%Y%m
        :param start_year_month:
        :param end_year_month:
        :return: [%Y%m]
        """
        month_list = pd.Series(
            pd.date_range(parse(start_year_month), parse(end_year_month), freq="m",)
        ).apply(lambda x: x.strftime("%Y-%m"))
        return month_list


if __name__ == "__main__":
    print(SpecialDatePoints.get_month_list("2020-02", "2020-08"))



import json

from lightning_fast.tools.json_tools.encoders.mongodb_doc_encoder import (
    MongodbDocEncoder,
)


class JsonEncoder:
    @classmethod
    def encode_mongodb(cls, mongodb_doc):
        return json.dumps(mongodb_doc, cls=MongodbDocEncoder)



import motor.motor_asyncio

from lightning_fast.mongodb.clients.client_descriptor.base.client_descriptor_base import ClientDescriptorBase


class AsyncMongodbClient(ClientDescriptorBase):
    def _create_client(self) -> motor.motor_asyncio.AsyncIOMotorClient:
        return motor.motor_asyncio.AsyncIOMotorClient(self.mongodb_url)



import os
import re
from functools import wraps

import luigi
from dateutil.parser import parse


"""
本页decorator务必配合task_bask使用，因为task_base有固定的文件名格式
"""


def delete_task_before_start_date(save_path_keywords: str, protect_end_date: str):
    """
    根据start_date字段，删除保护日期之后，当前之前的输出文件。task路径中
    luigi.Task继承类中必须含有start_date与end_date.
    文件名eg. SpeechRecognitionEvaluationTask(target_type_target_text_start_date_2020-11-02).csv
    :param save_path_keywords 保存路径（文件夹目录，不是文件路径）中含有的特殊字符串，防止误删除
    :param protect_end_date 保护日期（日期格式可以parse即可），保护日期与其之前的文件不会被删除
    """

    def decorate(run):
        @wraps(run)
        def wrapper(self: luigi.Task):
            try:
                start_date = getattr(self, "start_date")
            except AttributeError:
                raise AttributeError("Must had start_date attribute")
            parent_path = os.path.abspath(os.path.join(self.output().path, os.pardir))
            if save_path_keywords not in parent_path:
                raise ValueError("欲删除目录有误，请检查代码以防误删文件")
            if not os.path.exists(parent_path):
                os.makedirs(parent_path)
            for file in os.listdir(parent_path):
                valid_file = True
                if file.startswith("."):
                    valid_file = False
                if not valid_file:
                    continue
                # eg. SpeechRecognitionEvaluationTask(target_type_target_text_start_date_2020-11-02).csv
                start_date_raw = re.search(r"(start_date_.*?)[)_]", file).group(1)
                if not start_date_raw:
                    continue
                else:
                    start_date_str = start_date_raw.split("_")[-1]
                end_date_raw = re.search(r"(end_date_.*?)[)_]", file).group(1)
                if not end_date_raw:
                    continue
                else:
                    end_date_str = end_date_raw.split("_")[-1]
                try:
                    parsed_start_date = parse(start_date_str)
                    parsed_end_date = parse(end_date_str)
                except ValueError:
                    continue
                if parsed_end_date <= parse(protect_end_date):
                    continue
                else:
                    if parsed_end_date < parse(
                        str(start_date)
                    ) and parsed_start_date < parse(str(start_date)):
                        os.remove(os.path.join(parent_path, file))
            return run(self)

        return wrapper

    return decorate


def delete_task_before_date(save_path_keywords: str, protect_end_date: str):
    """
    根据date字段，删除保护日期之后，当前之前的输出文件。task路径中
    luigi.Task继承类中必须含有date.
    文件名eg. SpeechRecognitionEvaluationTask(target_type_target_text_date_2020-11-02).csv
    :param save_path_keywords 保存路径（文件夹目录，不是文件路径）中含有的特殊字符串，防止误删除
    :param protect_end_date 保护日期（日期格式可以parse即可），保护日期与其之前的文件不会被删除
    """

    def decorate(run):
        @wraps(run)
        def wrapper(self: luigi.Task):
            try:
                date = getattr(self, "date")
            except AttributeError:
                raise AttributeError("Must had date attribute")
            parent_path = os.path.abspath(os.path.join(self.output().path, os.pardir))
            if save_path_keywords not in parent_path:
                raise ValueError("欲删除目录有误，请检查代码以防误删文件")
            if not os.path.exists(parent_path):
                os.makedirs(parent_path)
            for file in os.listdir(parent_path):
                valid_file = True
                for name, parameter in self.param_kwargs.items():
                    if name != "date" and str(parameter) not in file:
                        valid_file = False
                if not valid_file:
                    continue
                # eg. SpeechRecognitionEvaluationTask(target_type_target_text_date_2020-11-02).csv
                date_raw = re.search(r"(date_.*?)[)_]", file).group(1)
                if not date_raw:
                    continue
                else:
                    date_str = date_raw.split("_")[-1]
                try:
                    parsed_date = parse(date_str)
                except ValueError:
                    continue
                if parsed_date <= parse(protect_end_date):
                    continue
                if parsed_date < parse(str(date)):
                    os.remove(os.path.join(parent_path, file))
            return run(self)

        return wrapper

    return decorate



from redis import ConnectionPool

from lightning_fast.redis_connection.pool.base.pool_descriptor_base import (
    PoolDescriptorBase,
)


class SyncRedisPool(PoolDescriptorBase):
    def _create_pool(self) -> ConnectionPool:
        return ConnectionPool.from_url(self.redis_url)



from lightning_fast.tools import (
    # mail_tools,
    # luigi_tools,
    logging_tools,
    serialize_tools,
    meta_tools,
    nest_object_tools,
    string_tools,
    path_tools,
    time_tools,
    json_tools,
    distributed_tools,
    # dask_tools,
)

__all__ = [
    # "mail_tools",
    # "luigi_tools",
    "logging_tools",
    "serialize_tools",
    "meta_tools",
    "nest_object_tools",
    "string_tools",
    "path_tools",
    "time_tools",
    "json_tools",
    "distributed_tools"
    # "dask_tools",
]



import logging
import os
from typing import TypeVar, Generic

import yaml

T = TypeVar("T")


class Config(Generic[T]):
    """
    using BASE_DIR = pathlib.Path(__file__).absolute().parent define your base path
    using
    class DirectoryCollection(dict):
        base_dir = DirectoryDescriptor(BASE_DIR)
        data_dir = DirectoryDescriptor(DATA_DIR)
        log_dir = DirectoryDescriptor(LOG_DIR)
        tmp_dir = DirectoryDescriptor(TMP_DIR)
    define your directory collection
    """

    def __init__(self, config_path, directories: T):
        self.config_path = config_path
        self.directories: T = directories
        self.environment = None
        self.important_static_config = None
        self.mongodbs = None
        self.redis = None
        self.external_api = None
        self.common_static_config = None
        self.kafka_producer = None
        self._get_env()
        self._get_config()
        self._set_log_level()

    def _get_env(self):
        if (
            not os.environ.get("CODE_ENV")
            or os.environ.get("CODE_ENV") == "development"
        ):
            self.environment = "development"
        elif os.environ.get("CODE_ENV") == "staging":
            self.environment = "staging"
        elif os.environ.get("CODE_ENV") == "production":
            self.environment = "production"
        else:
            raise ValueError("Env Wrong!")
        print(f"Current 'CODE_ENV' is '{self.environment}'")

    def _set_log_level(self):
        if self.environment == "production":
            logging.basicConfig(level=logging.INFO)
        else:
            logging.basicConfig(level=logging.DEBUG)

    def _get_config(self):

        if os.path.exists(self.config_path):
            with open(self.config_path) as f:
                static_config = yaml.safe_load(f)
                if static_config is None:
                    self.important_static_config = None
                else:
                    if self.environment == "development":
                        self.important_static_config = static_config.get("development")
                    elif self.environment == "staging":
                        self.important_static_config = static_config.get("staging")
                    elif self.environment == "production":
                        self.important_static_config = static_config.get("production")
                    else:
                        raise ValueError("Env Wrong!")
                    if not self.important_static_config:
                        logging.warning(
                            "Not have import static config (development or production)"
                        )
                    else:
                        if "mongodbs" not in self.important_static_config:
                            logging.warning("Not have 'mongodbs' in settings.yaml")
                            self.mongodbs = None
                        else:
                            self.mongodbs = self.important_static_config["mongodbs"]
                        if "redis" not in self.important_static_config:
                            logging.warning("Not have 'redis' in settings.yaml")
                            self.redis = None
                        else:
                            self.redis = self.important_static_config["redis"]
                        if "external_api" not in self.important_static_config:
                            logging.warning("Not have 'external_api' in settings.yaml")
                            self.external_api = None
                        else:
                            self.external_api = self.important_static_config[
                                "external_api"
                            ]
                self.common_static_config = static_config.get("common")



from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import List, Dict, Optional

import aiosmtplib

from lightning_fast.tools.mail_tools.smtp_base import SmtpBase


class AioSmtpServer(SmtpBase):
    """
    使用此类可以异步方式发送邮件
    """

    def __init__(self, host: str, email_from: str):
        self.host = host
        self.email_from = email_from

    async def send_email_to_one(
        self, to: str, subject: str, title: str, content: str, to_name: str = None
    ) -> None:
        """
        :param to: 邮件发送至地址
        :param subject: 邮件主题
        :param title: 邮件内容中标题
        :param content: 邮件内容
        :param to_name: 邮件发送至名字
        :return: 无返回
        """
        # await smtp.login(self.email_from, None)
        smtp = aiosmtplib.SMTP(hostname=self.host, port=25)
        async with smtp:
            message = MIMEMultipart("mixed")
            msg = MIMEText(
                self.get_common_html_email_content(title, content, to, to_name),
                "html",
                "utf-8",
            )
            message["From"] = self.email_from
            message["To"] = to
            message["Subject"] = subject
            message.attach(msg)
            await smtp.send_message(message)

    async def send_email_to_many(
        self,
        tos: List[str],
        subject: str,
        title: str,
        content: str,
        from_name: Optional[str] = None,
        bytes_mapping: Optional[Dict[str, bytes]] = None,
    ) -> None:
        """
        :param tos: 邮件发送至地址列表
        :param subject: 邮件主题
        :param title: 邮件内容中标题
        :param content: 邮件内容
        :param from_name: 邮件发送人名字
        :param bytes_mapping: 文件名与文件bytes对应的mapping
        :return: 无返回
        """
        # await smtp.login(self.email_from, None)
        if not bytes_mapping:
            bytes_mapping = {}
        smtp = aiosmtplib.SMTP(hostname=self.host, port=25)
        async with smtp:
            message = MIMEMultipart("mixed")
            msg = MIMEText(
                self.get_html_content_with_hello(title, content, "您好", from_name),
                "html",
                "utf-8",
            )
            message["From"] = self.email_from
            message["To"] = ",".join(tos)
            message["Subject"] = subject
            message.attach(msg)
            for name, bytes_like in bytes_mapping.items():
                part = MIMEApplication(bytes_like, Name=name)
                part["Content-Disposition"] = 'attachment; filename="%s"' % name
                message.attach(part)
            await smtp.send_message(message, recipients=tos)

    async def send_email_to_team_leader_with_cc(
        self,
        to: str,
        cc: list,
        subject: str,
        title: str,
        content: str,
        from_name: str = None,
    ) -> None:
        """
        :param to: 邮件发送至地址
        :param cc: 抄送地址
        :param subject: 邮件主题
        :param title: 邮件内容中标题
        :param content: 邮件内容
        :param from_name: 邮件发送人名字
        :return: 无返回
        """
        # await smtp.login(self.email_from, None)
        smtp = aiosmtplib.SMTP(hostname=self.host, port=25)
        async with smtp:
            message = MIMEMultipart("mixed")
            msg = MIMEText(
                self.get_html_content_with_hello(title, content, "您好", from_name),
                "html",
                "utf-8",
            )
            message["From"] = self.email_from
            message["To"] = to
            message["Cc"] = ",".join(cc)
            message["Subject"] = subject
            message.attach(msg)
            await smtp.send_message(message)



import unittest
from utils.local_c_tools_compiler import LocalCToolsCompiler


class TestLabelEncoder(unittest.TestCase):

    test_strings = ["中文1", "中文2", "中文2", "符号，", "符号。", "符号；", "", "english"]
    test_encode_result = [0, 1, 1, 2, 3, 4, 5, 6]

    def setUp(self) -> None:
        LocalCToolsCompiler().compile_local_c_tools()

    def test_encoder_fit(self):
        # 测试是否能正常生成encoder
        encoders = LocalCToolsCompiler.import_encoders()
        label_encoder = encoders.LabelEncoder()
        one_d_string_vector = encoders.OneDStringVector(self.test_strings)
        label_encoder.encode_1d(one_d_string_vector, 2)
        print(label_encoder)

    def test_encoder_transform(self):
        # 测试是否能正常transform
        encoders = LocalCToolsCompiler.import_encoders()
        label_encoder = encoders.LabelEncoder()
        one_d_string_vector = encoders.OneDStringVector(self.test_strings)
        label_encoder.encode_1d(one_d_string_vector, 2)
        transform_result = label_encoder.transform_1d(one_d_string_vector, 2)
        self.assertListEqual(self.test_encode_result, list(transform_result))


if __name__ == "__main__":
    unittest.main()



import json
from bson import json_util


class MongodbDocDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self,
            object_hook=json_util.object_hook,
            object_pairs_hook=json_util.object_pairs_hook,
            *args,
            **kwargs
        )



from typing import ClassVar, Dict, Iterable, Iterator, List

from typing import overload

class LabelEncoder:
    def __init__(self) -> None: ...
    def encode_1d(self, arg0: OneDStringVector) -> None: ...
    def transform_1d(self, arg0: OneDStringVector, arg1: int) -> List[int]: ...
    @property
    def label_map(self) -> Dict[str,int]: ...

class OneDStringVector:
    __hash__: ClassVar[None] = ...
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, arg0: OneDStringVector) -> None: ...
    @overload
    def __init__(self, arg0: Iterable) -> None: ...
    def append(self, x: str) -> None: ...
    def clear(self) -> None: ...
    def count(self, x: str) -> int: ...
    @overload
    def extend(self, L: OneDStringVector) -> None: ...
    @overload
    def extend(self, L: Iterable) -> None: ...
    def insert(self, i: int, x: str) -> None: ...
    @overload
    def pop(self) -> str: ...
    @overload
    def pop(self, i: int) -> str: ...
    def remove(self, x: str) -> None: ...
    def __bool__(self) -> bool: ...
    def __contains__(self, x: str) -> bool: ...
    @overload
    def __delitem__(self, arg0: int) -> None: ...
    @overload
    def __delitem__(self, arg0: slice) -> None: ...
    def __eq__(self, arg0: OneDStringVector) -> bool: ...
    @overload
    def __getitem__(self, s: slice) -> OneDStringVector: ...
    @overload
    def __getitem__(self, arg0: int) -> str: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...
    def __ne__(self, arg0: OneDStringVector) -> bool: ...
    @overload
    def __setitem__(self, arg0: int, arg1: str) -> None: ...
    @overload
    def __setitem__(self, arg0: slice, arg1: OneDStringVector) -> None: ...



from lightning_fast.tools.luigi_tools.task_base import TaskBase
from lightning_fast.tools.luigi_tools.decorator import (
    delete_task_before_start_date,
    delete_task_before_date,
)
from lightning_fast.tools.luigi_tools.hive_dt_partition_check import (
    HiveDtPartitionCheck,
)


__all__ = [
    "TaskBase",
    "delete_task_before_date",
    "delete_task_before_start_date",
    "HiveDtPartitionCheck",
]



from typing import Callable


class NameConverter:
    """
    这是一个把嵌套dict、list格式的对象所有key进行转换的类，常用为将一个文档转为驼峰传入前端。
    传入函数可以在convention_functions中找到
    """
    @classmethod
    def change_naming_convention(cls, origin, convert_function: Callable):
        if isinstance(origin, list):
            new_list = []
            for info in origin:
                new_list.append(cls.change_naming_convention(info, convert_function))
            return new_list
        elif isinstance(origin, dict):
            new_dict = {}
            for k, v in origin.items():
                # 这里是进行特殊key特殊处理的地方。如果遇到需要这样处理的情况再改进。
                # if k == "answers":
                #     for sub_v in v:
                #         sub_v["subAnswers"] = []
                if isinstance(v, dict):
                    new_v = cls.change_naming_convention(v, convert_function)
                elif isinstance(v, list):
                    new_v = [
                        cls.change_naming_convention(x, convert_function) for x in v
                    ]
                else:
                    new_v = v
                new_dict[convert_function(k)] = new_v
            return new_dict
        else:
            return origin



from lightning_fast.encoders.c_encoders.c_label_encoder import CLabelEncoder
from lightning_fast.encoders.c_encoders.c_utils import DataConverter
from lightning_fast.encoders.common_encoders.common_label_encoder import (
    CommonLabelEncoder,
)


__all__ = ["CLabelEncoder", "DataConverter", "CommonLabelEncoder"]



from datetime import date


class SmtpBase:
    @classmethod
    def get_common_html_email_content(
        cls,
        title: str,
        content: str,
        to: str,
        from_name: str = None,
        to_name: str = None,
    ) -> str:
        """
        :param title: 邮件正文中的标题
        :param content: 邮件内容
        :param to: 邮件发送至地址
        :param from_name: 邮件发送名字
        :param to_name: 邮件发送至名字
        :return: html版内容
        """
        to_name = to_name if to_name else to
        time_string = date.today()
        email_content = f"""
            <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" 
            "https://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
            <html xmlns="https://www.w3.org/1999/xhtml">
                <head>
                    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
                    <title>{title}</title>
                    <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
                </head>
                <body style="margin: 0; padding: 0;">
                    <div>Hi {to_name},</div>
                    <p>{content}</p>
                    <div style="position: fixed; right: 0; bottom: 0;">
                        <div>{'统计服务' if not from_name else from_name}</div>
                        <div>{time_string}</div>
                    </div>
                </body>
            </html>
        """
        return email_content

    @classmethod
    def get_html_content_with_hello(
        cls,
        title: str,
        content: str,
        hello: str,
        from_name: str,
    ) -> str:
        """
        :param title: 邮件正文中的标题
        :param content: 邮件内容
        :param hello: 邮件第一行
        :param from_name: 邮件发送人名字
        :return: html版内容
        """
        time_string = date.today()
        email_content = f"""
            <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" 
            "https://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
            <html xmlns="https://www.w3.org/1999/xhtml">
                <head>
                    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
                    <title>{title}</title>
                    <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
                </head>
                <body style="margin: 0; padding: 0;">
                    <div>{hello},</div>
                    <p>{content}</p>
                    <div style="position: fixed; right: 0; bottom: 0;">
                        <div>{from_name}</div>
                        <div>{time_string}</div>
                    </div>
                </body>
            </html>
        """
        return email_content



import datetime
import json

from bson import json_util


class MongodbDocEncoder(json.JSONEncoder):
    """
    将json格式的题目中不符合规范的数值类型进行转换
    """

    def default(self, o):
        return json_util.default(o)



import os
import pickle

N_BYTES = 2 ** 31
MAX_BYTES = 2 ** 31 - 1


class PickleObject(object):
    """
    这是一个pickle的补充类, 主要用来处理类型过大的时候, 可以用这个类来分批次序列化储存与读取
    """

    @classmethod
    def write_to_file(cls, file_path, data):
        bytes_out = pickle.dumps(data)
        with open(file_path, "wb") as f_out:
            for idx in range(0, len(bytes_out), MAX_BYTES):
                f_out.write(bytes_out[idx: idx + MAX_BYTES])

    @classmethod
    def read_from_file(cls, file_path):
        bytes_in = bytearray(0)
        input_size = os.path.getsize(file_path)
        with open(file_path, "rb") as f_in:
            for _ in range(0, input_size, MAX_BYTES):
                bytes_in += f_in.read(MAX_BYTES)
        data = pickle.loads(bytes_in)
        return data



import smtplib
from email.header import Header
from email.mime.text import MIMEText

from lightning_fast.tools.mail_tools.smtp_base import SmtpBase


class SyncSmtpServer(SmtpBase):
    def __init__(self, host, email_from, password=None):
        self.host = host
        self.email_from = email_from
        self.password = password

    def _init_server(self):
        server = smtplib.SMTP(self.host)
        if self.email_from and self.password:
            server.login(self.email_from, self.password)
        return server

    def send_email_to_one(
        self, to: str, subject: str, title: str, content: str, to_name: str = None
    ) -> None:
        html_content = self.get_common_html_email_content(title, content, to, to_name)
        msg = MIMEText(html_content, _subtype="html")
        server = self._init_server()
        msg["From"] = self.email_from
        msg["To"] = to
        msg["Subject"] = Header(subject)
        server.sendmail(self.email_from, to, msg.as_string())
        server.close()



from io import StringIO
from typing import Optional, Union

import pandas as pd
import dask.dataframe as dd


class DaskReader:
    @classmethod
    def read_raw_hive_data(
        cls, file_path, sep="\t", dtype=None, partitions: int = 1, limit: Optional[int] = None
    ) -> Union[pd.DataFrame, dd.DataFrame]:
        if limit is None:
            df = dd.read_csv(file_path, sep=sep, dtype=dtype)
        else:
            df_string = ""
            with open(file_path) as f:
                current_count = 0
                while current_count < limit:
                    df_string += f.readline()
                    current_count += 1
            df = dd.from_pandas(
                pd.read_csv(StringIO(df_string), sep=sep, dtype=dtype),
                npartitions=partitions,
            )
        return df



from lightning_fast.config_factory.config_constructor import Config
from lightning_fast.config_factory.yaml_example import YamlExample
from lightning_fast.config_factory.simple_directory_collection import SimpleDirectoryCollection


__all__ = ["Config", "YamlExample", "SimpleDirectoryCollection"]



class SingletonMeta(type):
    """
    使用此类作为metaclass则类会变成单例
    此类根据类的名字、实例化时传入的参数进行去重。即如果传入不同的参数则会形成不同的单例
    """

    _instances = {}

    def __call__(cls, *args, **kwargs) -> object:
        # kwd_mark = object()
        # key = args + (kwd_mark, cls.__name__) + tuple(sorted(kwargs.items()))
        key = str(args) + str((cls.__name__,)) + str(tuple(sorted(kwargs.items())))
        if key not in cls._instances:
            cls._instances[key] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[key]

    @classmethod
    def clear(mcs):
        mcs._instances = {}



from abc import abstractmethod
import pathlib
import re
import sys
from typing import Optional

import luigi


class TaskBase(luigi.Task):
    """
    封装luigi.Task。
    使用类相同路径在data目录下创建生成文件路径
    使用task名字与参数构建生成文件名字。
    如果想自定义文件后缀，则类中覆盖output_file_type
    此Base自行继承并复写base_dir与data_dir, 否则不可用
    """

    output_file_type = "csv"

    @property
    @abstractmethod
    def base_dir(self) -> Optional[pathlib.Path]:
        """
        返回当前项目根目录绝对路径
        :return: 项目根目录绝对路径
        """
        return None

    @property
    @abstractmethod
    def data_dir(self) -> Optional[pathlib.Path]:
        """
        返回欲保存的数据目录的绝对路径
        :return: 保存数据目录绝对路径
        """
        return None

    @property
    def environment(self) -> str:
        """
        返回环境字符串，需要改变则复写此方法
        :return: 环境字符串，比如development
        """
        return "development"

    def output_directory(self) -> Optional[pathlib.Path]:
        """
        有趣的机制，如果这里使用类方法，那么cls.__class__会指向元类。
        这是目前发现的唯一的一个类方法导致与直觉相反的情况。
        """
        if not self.base_dir or not self.data_dir:
            raise ValueError("Please imply base_dir and data_dir")
        file_type_suffix_re = re.compile(r"\.[^.\\/]*?$")
        if self.__class__ == TaskBase:
            return None
        data_dir = self.data_dir
        current_path = sys.modules[self.__class__.__module__].__file__
        current_file_path = pathlib.Path(current_path).absolute()
        current_data_dir = data_dir / current_file_path.relative_to(self.base_dir)
        current_data_dir = pathlib.Path(
            file_type_suffix_re.sub("", str(current_data_dir))
        )
        current_data_dir.mkdir(parents=True, exist_ok=True)
        return current_data_dir

    def output(self):
        output_file_name = self.__class__.__name__
        for parameter_name, parameter_value in self.to_str_params().items():
            output_file_name += "(" + parameter_name + "_" + parameter_value + ")"
        output_file_name += (
            f"(environment_{self.environment})" + f".{self.output_file_type}"
        )
        return luigi.LocalTarget(self.output_directory() / output_file_name)



from ctypes import create_string_buffer


class DataConverter(object):
    @classmethod
    def string_list_to_c_string(cls, string_list):
        c_string = create_string_buffer("\t".join(string_list).encode("utf-8"))
        return c_string
